"""
dexml: A dead-simple Object-XML mapper for Python
"""
from dexml import fields
from dexml.constants import Status
from dexml.exceptions import ParseError, RenderError, XmlError
from dexml.model import Model

__version__ = "0.1.1"


__all__ = [
    "Model",
    "fields",
    "ParseError",
    "RenderError",
    "XmlError",
    "Status",
]



from typing import List
from xml.dom import minidom

from dexml import constants, exceptions
from dexml.fields import field as dexml_field
from dexml.meta import Meta, ModelMetaclass


class Model(metaclass=ModelMetaclass):
    """Base class for dexml Model objects.

    Subclasses of Model represent a concrete type of object that can parsed
    from or rendered to an XML document.  The mapping to/from XML is controlled
    by two things:

        * attributes declared on an inner class named 'meta'
        * fields declared using instances of fields.Field

    Here's a quick example:

        class Person(dexml.Model):
            # This overrides the default tagname of 'Person'
            class meta
                tagname = "person"
            # This maps to a 'name' attributr on the <person> tag
            name = fields.String()
            # This maps to an <age> tag within the <person> tag
            age = fields.Integer(tagname='age')

    See the 'Meta' class in this module for available meta options, and the
    'fields' submodule for available field types.
    """

    _fields: List[dexml_field.Field] = []
    meta: Meta

    def __init__(self, **kwds):
        """Default Model constructor.

        Keyword arguments that correspond to declared fields are processed
        and assigned to that field.
        """
        for field in self._fields:
            try:
                setattr(self, field.field_name, kwds[field.field_name])
            except KeyError:
                pass

    @classmethod
    def parse(cls, xml, tagname=None):
        """Produce an instance of this model from some xml.

        The given xml can be a string, a readable file-like object, or
        a DOM node; we might add support for more types in the future.
        """
        self = cls()
        node = self._make_xml_node(xml)
        self.validate_xml_node(node, tagname)
        #  Keep track of fields that have successfully parsed something
        fields_found = []
        #  Try to consume all the node's attributes
        attrs = node.attributes.values()
        for field in self._fields:
            unused_attrs = field.parse_attributes(self, attrs)
            if len(unused_attrs) < len(attrs):
                fields_found.append(field)
            attrs = unused_attrs
        for attr in attrs:
            self._handle_unparsed_node(attr)
        #  Try to consume all child nodes
        if self.meta.order_sensitive:
            self._parse_children_ordered(node, self._fields, fields_found)
        else:
            self._parse_children_unordered(node, self._fields, fields_found)
        #  Check that all required fields have been found
        for field in self._fields:
            if field.required and field not in fields_found:
                err = "required field not found: '%s'" % (field.field_name,)
                raise exceptions.ParseError(err)
            field.parse_done(self)
        #  All done, return the instance so created
        return self

    def _parse_children_ordered(self, node, fields, fields_found):
        """Parse the children of the given node using strict field ordering."""
        cur_field_idx = 0
        for child in node.childNodes:
            idx = cur_field_idx
            #  If we successfully break out of this loop, one of our
            #  fields has consumed the node.
            while idx < len(fields):
                field = fields[idx]
                res = field.parse_child_node(self, child)
                if res is constants.Status.PARSE_DONE:
                    if field not in fields_found:
                        fields_found.append(field)
                    cur_field_idx = idx + 1
                    break
                if res is constants.Status.PARSE_MORE:
                    if field not in fields_found:
                        fields_found.append(field)
                    cur_field_idx = idx
                    break
                if res is constants.Status.PARSE_CHILDREN:
                    if field not in fields_found:
                        fields_found.append(field)
                    self._parse_children_ordered(child, [field], fields_found)
                    cur_field_idx = idx
                    break
                idx += 1
            else:
                self._handle_unparsed_node(child)

    def _parse_children_unordered(self, node, fields, fields_found):
        """Parse the children of the given node using loose field ordering."""
        done_fields = {}
        for child in node.childNodes:
            idx = 0
            #  If we successfully break out of this loop, one of our
            #  fields has consumed the node.
            while idx < len(fields):
                if idx in done_fields:
                    idx += 1
                    continue
                field = fields[idx]
                res = field.parse_child_node(self, child)
                if res is constants.Status.PARSE_DONE:
                    done_fields[idx] = True
                    if field not in fields_found:
                        fields_found.append(field)
                    break
                if res is constants.Status.PARSE_MORE:
                    if field not in fields_found:
                        fields_found.append(field)
                    break
                if res is constants.Status.PARSE_CHILDREN:
                    if field not in fields_found:
                        fields_found.append(field)
                    self._parse_children_unordered(child, [field], fields_found)
                    break
                idx += 1
            else:
                self._handle_unparsed_node(child)

    def _handle_unparsed_node(self, node):
        if not self.meta.ignore_unknown_elements:
            if node.nodeType == node.ELEMENT_NODE:
                err = "unknown element: %s" % (node.nodeName,)
                raise exceptions.ParseError(err)
            if node.nodeType in (node.TEXT_NODE, node.CDATA_SECTION_NODE):
                if node.nodeValue.strip():
                    err = "unparsed text node: %s" % (node.nodeValue,)
                    raise exceptions.ParseError(err)
            if node.nodeType == node.ATTRIBUTE_NODE:
                if not node.nodeName.startswith("xml"):
                    err = "unknown attribute: %s" % (node.name,)
                    raise exceptions.ParseError(err)

    def render(self, encoding=None, fragment=False, pretty=False, nsmap=None):
        """Produce XML from this model's instance data.

        A unicode string will be returned if any of the objects contain
        unicode values; specifying the 'encoding' argument forces generation
        of a bytestring.

        By default a complete XML document is produced, including the
        leading "<?xml>" declaration.  To generate an XML fragment set
        the 'fragment' argument to True.
        """
        if nsmap is None:
            nsmap = {}
        data = []
        header = '<?xml version="1.0" ?>'
        if encoding:
            header = '<?xml version="1.0" encoding="%s" ?>' % (encoding,)
        if not fragment:
            data.append(header)

        data.extend(self._render(nsmap))
        xml = "".join(data)
        if pretty:
            xml = minidom.parseString(xml).toprettyxml()
            # Hack for removing the `<?xml version="1.0"?>` header that
            # minidom adds when pretty printing.
            line_break_position = xml.find("\n") + 1
            headless_xml = xml[line_break_position:]
            if fragment:
                xml = headless_xml
            elif encoding:
                # Minidom also removes the header (or just the `encoding` key)
                # if it is present
                xml = header + "\n" + headless_xml
        if encoding:
            xml = xml.encode(encoding)
        return xml

    def irender(self, encoding=None, fragment=False, nsmap=None):
        """Generator producing XML from this model's instance data.

        If any of the objects contain unicode values, the resulting output
        stream will be a mix of bytestrings and unicode; specify the 'encoding'
        arugment to force generation of bytestrings.

        By default a complete XML document is produced, including the
        leading "<?xml>" declaration.  To generate an XML fragment set
        the 'fragment' argument to True.
        """
        if nsmap is None:
            nsmap = {}
        if not fragment:
            if encoding:
                decl = f'<?xml version="1.0" encoding="{encoding}" ?>'
                yield decl.encode(encoding)
            else:
                yield '<?xml version="1.0" ?>'
        if encoding:
            for data in self._render(nsmap):
                if isinstance(data, str):
                    data = data.encode(encoding)
                yield data
        else:
            for data in self._render(nsmap):
                yield data

    def _render(self, nsmap, tagname=None):
        """Generator rendering this model as an XML fragment."""
        #  Determine opening and closing tags
        pushed_ns = False
        if not tagname:
            tagname = self.meta.tagname
        if self.meta.namespace:
            namespace = self.meta.namespace
            prefix = self.meta.namespace_prefix
            try:
                cur_ns = nsmap[prefix]
            except KeyError:
                cur_ns = []
                nsmap[prefix] = cur_ns
            if prefix:
                tagname = f"{prefix}:{tagname}"
                open_tag_contents = [tagname]
                if not cur_ns or cur_ns[0] != namespace:
                    cur_ns.insert(0, namespace)
                    pushed_ns = True
                    open_tag_contents.append(f'xmlns:{prefix}="{namespace}"')
                close_tag_contents = tagname
            else:
                open_tag_contents = [tagname]
                if not cur_ns or cur_ns[0] != namespace:
                    cur_ns.insert(0, namespace)
                    pushed_ns = True
                    open_tag_contents.append(f'xmlns="{namespace}"')
                close_tag_contents = tagname
        else:
            open_tag_contents = [tagname]
            close_tag_contents = tagname
        used_fields = set()
        open_tag_contents.extend(self._render_attributes(used_fields, nsmap))
        #  Render each child node
        children = self._render_children(used_fields, nsmap)
        try:
            first_child = next(children)
        except StopIteration:
            yield f'<{" ".join(open_tag_contents)} />'
        else:
            yield f'<{" ".join(open_tag_contents)}>'
            yield first_child
            for child in children:
                yield child
            yield f"</{close_tag_contents}>"
        #  Check that all required fields actually rendered something
        for field in self._fields:
            if field.required and field not in used_fields:
                raise exceptions.RenderError(
                    "Field '%s' is missing" % (field.field_name,)
                )
        #  Clean up
        if pushed_ns:
            nsmap[prefix].pop(0)

    def _render_attributes(self, used_fields, nsmap):
        for field in self._fields:
            val = getattr(self, field.field_name)
            datas = iter(field.render_attributes(self, val, nsmap))
            try:
                data = next(datas)
            except StopIteration:
                pass
            else:
                used_fields.add(field)
                yield data
                for data in datas:
                    yield data

    def _render_children(self, used_fields, nsmap):
        for field in self._fields:
            val = getattr(self, field.field_name)
            datas = iter(field.render_children(self, val, nsmap))
            try:
                data = next(datas)
            except StopIteration:
                pass
            else:
                used_fields.add(field)
                yield data
                for data in datas:
                    yield data

    @staticmethod
    def _make_xml_node(xml):
        """Transform a variety of input formats to an XML DOM node."""
        try:
            ntype = xml.nodeType
        except AttributeError as exc:
            if isinstance(xml, bytes):
                try:
                    xml = minidom.parseString(xml)
                except Exception as exception:
                    raise exceptions.XmlError(exception)
            elif isinstance(xml, str):
                try:
                    #  Try to grab the "encoding" attribute from the XML.
                    #  It probably won't exist, so default to utf8.
                    encoding = constants.XML_ENCODING_RE.match(xml)
                    if encoding is None:
                        encoding = "utf8"
                    else:
                        encoding = encoding.group(1)
                    xml = minidom.parseString(xml.encode(encoding))
                except Exception as exception:
                    raise exceptions.XmlError(exception)
            elif hasattr(xml, "read"):
                try:
                    xml = minidom.parse(xml)
                except Exception as exception:
                    raise exceptions.XmlError(exception)
            else:
                raise ValueError("Can't convert that to an XML DOM node") from exc
            node = xml.documentElement
        else:
            if ntype == xml.DOCUMENT_NODE:
                node = xml.documentElement
            else:
                node = xml
        return node

    @classmethod
    def validate_xml_node(cls, node, tagname=None):
        """Check that the given xml node is valid for this object.

        Here 'valid' means that it is the right tag, in the right
        namespace.  We might add more eventually...
        """
        if node.nodeType != node.ELEMENT_NODE:
            err = "Class '%s' got a non-element node"
            err = err % (cls.__name__,)
            raise exceptions.ParseError(err)
        if cls.meta.case_sensitive:
            if not (cls.meta.tagname and node.localName == cls.meta.tagname) and not (
                tagname and tagname == node.localName
            ):
                err = "Class '%s' got tag '%s' (expected '%s')"
                err = err % (cls.__name__, node.localName, cls.meta.tagname)
                raise exceptions.ParseError(err)
        else:
            if not (node.localName.lower() == cls.meta.tagname.lower()) and not (
                tagname and tagname.lower() == node.localName.lower()
            ):
                err = "Class '%s' got tag '%s' (expected '%s')"
                err = err % (cls.__name__, node.localName, cls.meta.tagname)
                raise exceptions.ParseError(err)
        if cls.meta.namespace:
            if node.namespaceURI != cls.meta.namespace:
                err = "Class '%s' got namespace '%s' (expected '%s')"
                err = err % (cls.__name__, node.namespaceURI, cls.meta.namespace)
                raise exceptions.ParseError(err)
        else:
            if node.namespaceURI:
                err = "Class '%s' got namespace '%s' (expected no namespace)"
                err = err % (
                    cls.__name__,
                    node.namespaceURI,
                )
                raise exceptions.ParseError(err)



import enum
import re


class Status(enum.Enum):
    """Constant returned by a Field when it has finished parsing."""

    PARSE_DONE: str = "done"
    PARSE_MORE: str = "more"
    PARSE_SKIP: str = "skip"
    PARSE_CHILDREN: str = "children"


#  You can use this re to extract the encoding declaration from the XML
#  document string.  Hopefully you won't have to, but you might need to...
XML_ENCODING_RE = re.compile(
    "<\\?xml [^>]*encoding=[\"']([a-zA-Z0-9\\.\\-\\_]+)[\"'][^>]*?>"
)



import random
from typing import Any, Optional
from xml.sax.saxutils import escape, quoteattr

from dexml import constants, exceptions
from dexml.fields import field


class Value(field.Field):
    """Field subclass that holds a simple scalar value.

    This field.Field subclass contains the common logic to parse/render simple
    scalar value fields - fields that don't required any recursive parsing.
    Individual subclasses should provide the parse_value() and render_value()
    methods to do type coercion of the value.

    Value fields can also have a default value, specified by the 'default'
    keyword argument.

    By default, the field maps to an attribute of the model's XML node with
    the same name as the field declaration.  Consider:

        class MyModel(Model):
            my_field = fields.Value(default="test")


    This corresponds to the XML fragment "<MyModel my_field='test' />".
    To use a different name specify the 'attrname' kwd argument.  To use
    a subtag instead of an attribute specify the 'tagname' kwd argument.

    Namespaced attributes or subtags are also supported, by specifying a
    (namespace,tagname) pair for 'attrname' or 'tagname' respectively.
    """

    default: Any

    class Arguments(field.Field.Arguments):
        tagname = None
        attrname = None
        default = None

    def __init__(self, **kwds: Any) -> None:
        super().__init__(**kwds)
        if self.default is not None:
            self.required = False

    def _get_attrname(self) -> Optional[str]:
        if self.__dict__["tagname"]:
            return None
        attrname = self.__dict__["attrname"]
        if not attrname:
            attrname = self.field_name
        return attrname

    def _set_attrname(self, attrname) -> None:
        self.__dict__["attrname"] = attrname

    attrname = property(_get_attrname, _set_attrname)

    def _get_tagname(self):
        if self.__dict__["attrname"]:
            return None
        tagname = self.__dict__["tagname"]
        if tagname and not isinstance(tagname, (str, tuple)):
            tagname = self.field_name
        return tagname

    def _set_tagname(self, tagname: str) -> None:
        self.__dict__["tagname"] = tagname

    tagname = property(_get_tagname, _set_tagname)

    def __get__(self, instance, owner=None):
        val = super().__get__(instance, owner)
        if val is None:
            return self.default
        return val

    def parse_attributes(self, obj, attrs):
        #  Bail out if we're attached to a subtag rather than an attr.
        if self.tagname:
            return attrs
        unused = []
        attrname = self.attrname
        if isinstance(attrname, str):
            ns = None
        else:
            (ns, attrname) = attrname
        for attr in attrs:
            if attr.localName == attrname:
                if attr.namespaceURI == ns:
                    self.__set__(obj, self.parse_value(attr.nodeValue))
                else:
                    unused.append(attr)
            else:
                unused.append(attr)
        return unused

    def parse_child_node(self, obj, node):
        if not self.tagname:
            return constants.Status.PARSE_SKIP
        if self.tagname == ".":
            node = node.parentNode
        else:
            if not self._check_tagname(node, self.tagname):
                return constants.Status.PARSE_SKIP
        vals = []
        #  Merge all text nodes into a single value
        for child in node.childNodes:
            if child.nodeType not in (child.TEXT_NODE, child.CDATA_SECTION_NODE):
                raise exceptions.ParseError("non-text value node")
            vals.append(child.nodeValue)
        self.__set__(obj, self.parse_value("".join(vals)))
        return constants.Status.PARSE_DONE

    def render_attributes(self, obj, val, nsmap):
        if val is not None and val is not self.default and self.attrname:
            qaval = quoteattr(self.render_value(val))
            if isinstance(self.attrname, str):
                yield f"{self.attrname}={qaval}"
            else:
                m_meta = self.model_class.meta
                (ns, nm) = self.attrname
                if ns == m_meta.namespace and m_meta.namespace_prefix:
                    prefix = m_meta.namespace_prefix
                    yield f"{prefix}:{nm}={qaval}"
                elif ns is None:
                    yield f"{nm}={qaval}"
                else:
                    for (p, n) in iter(nsmap.items()):
                        if ns == n[0]:
                            prefix = p
                            break
                    else:
                        prefix = "p" + str(random.randint(0, 10000))
                        while prefix in nsmap:
                            prefix = "p" + str(random.randint(0, 10000))
                        yield f'xmlns:{prefix}="{ns}"'
                    yield f"{prefix}:{nm}={qaval}"

    def render_children(self, obj, val, nsmap):
        if val is not None and val is not self.default and self.tagname:
            val = self._esc_render_value(val)
            if self.tagname == ".":
                yield val
            else:
                attrs = ""
                #  By default, tag values inherit the namespace of their
                #  containing model class.
                if isinstance(self.tagname, str):
                    prefix = self.model_class.meta.namespace_prefix
                    localName = self.tagname
                else:
                    m_meta = self.model_class.meta
                    (ns, localName) = self.tagname
                    if not ns:
                        #  If we have an explicitly un-namespaced tag,
                        #  we need to be careful.  The model tag might have
                        #  set the default namespace, which we need to undo.
                        prefix = None
                        if m_meta.namespace and not m_meta.namespace_prefix:
                            attrs = ' xmlns=""'
                    elif ns == m_meta.namespace:
                        prefix = m_meta.namespace_prefix
                    else:
                        for (p, n) in iter(nsmap.items()):
                            if ns == n[0]:
                                prefix = p
                                break
                        else:
                            prefix = "p" + str(random.randint(0, 10000))
                            while prefix in nsmap:
                                prefix = "p" + str(random.randint(0, 10000))
                            attrs = ' xmlns:%s="%s"' % (prefix, ns)
                yield self._render_tag(val, prefix, localName, attrs)

    def _render_tag(self, val, prefix, localName, attrs):
        if val:
            if prefix:
                return f"<{prefix}:{localName}{attrs}>{val}</{prefix}:{localName}>"
            return f"<{localName}{attrs}>{val}</{localName}>"
        if prefix:
            return f"<{prefix}:{localName}{attrs} />"
        return f"<{localName}{attrs} />"

    def parse_value(self, val):
        return val

    def render_value(self, val):
        if not isinstance(val, str):
            val = str(val)
        return val

    def _esc_render_value(self, val):
        return escape(self.render_value(val))


class String(Value):
    """Field representing a simple string value."""

    # actually, the base Value() class will do this automatically.


class CDATA(Value):
    """String field rendered as CDATA."""

    def __init__(self, **kwds):
        super().__init__(**kwds)
        if self.__dict__.get("tagname", None) is None:
            raise ValueError("CDATA fields must have a tagname")

    def _esc_render_value(self, val):
        val = self.render_value(val)
        val = val.replace("]]>", "]]]]><![CDATA[>")
        return "<![CDATA[" + val + "]]>"


class Integer(Value):
    """Field representing a simple integer value."""

    def parse_value(self, val):
        return int(val)


class Float(Value):
    """Field representing a simple float value."""

    def parse_value(self, val):
        return float(val)


class Boolean(Value):
    """Field representing a simple boolean value.

    The strings corresponding to false are 'no', 'off', 'false' and '0',
    compared case-insensitively.  Note that this means an empty tag or
    attribute is considered True - this is usually what you want, since
    a completely missing attribute or tag can be interpreted as False.

    To enforce that the presence of a tag indicates True and the absence of
    a tag indicates False, pass the keyword argument "empty_only".
    """

    empty_only: bool

    class Arguments(Value.Arguments):
        empty_only = False

    def __init__(self, **kwds):
        super().__init__(**kwds)
        if self.empty_only:
            self.required = False

    def __set__(self, instance, value):
        instance.__dict__[self.field_name] = bool(value)

    def parse_value(self, val):
        if self.empty_only and val != "":
            raise ValueError("non-empty value in empty_only Boolean")
        if val.lower() in ("no", "off", "false", "0"):
            return False
        return True

    def render_children(self, obj, val, nsmap):
        if not val and self.empty_only:
            return []
        return super().render_children(obj, val, nsmap)

    def render_attributes(self, obj, val, nsmap):
        if not val and self.empty_only:
            return []
        return super().render_attributes(obj, val, nsmap)

    def render_value(self, val):
        if not val:
            return "false"
        if self.empty_only:
            return ""
        return "true"



from typing import List

from dexml import constants, exceptions
from dexml.fields import field as dexml_field
from dexml.fields import model as model_field


class Choice(dexml_field.Field):
    """Field subclass accepting any one of a given set of model_field.Model fields."""

    fields: List[dexml_field.Field]

    class Arguments(dexml_field.Field.Arguments):
        fields: List[dexml_field.Field] = []

    def __init__(self, *fields, **kwds):
        real_fields = []
        for field in fields:
            if isinstance(field, model_field.Model):
                real_fields.append(field)
            elif isinstance(field, str):
                real_fields.append(model_field.Model(field))
            else:
                raise ValueError("only Model fields are allowed within a Choice field")
        kwds["fields"] = real_fields
        super().__init__(**kwds)

    def parse_child_node(self, obj, node):
        for field in self.fields:
            field.field_name = self.field_name
            field.model_class = self.model_class
            res = field.parse_child_node(obj, node)
            if res is constants.Status.PARSE_MORE:
                raise ValueError("items in a Choice cannot return PARSE_MORE")
            if res is constants.Status.PARSE_DONE:
                return constants.Status.PARSE_DONE
        else:
            return constants.Status.PARSE_SKIP

    def render_children(self, obj, item, nsmap):
        if item is None:
            if self.required:
                raise exceptions.RenderError(
                    f"Field '{self.field_name}': required field is missing"
                )
        else:
            for data in item._render(nsmap=nsmap):
                yield data



import copy
from typing import Any, Dict, Optional

from dexml.fields.field import Field


class Meta:
    """Class holding meta-information about a dexml.Model subclass.

    Each dexml.Model subclass has an attribute 'meta' which is an instance
    of this class.  That instance holds information about how the model
    corresponds to XML, such as its tagname, namespace, and error handling
    semantics.  You would not ordinarily create an instance of this class;
    instead let the ModelMetaclass create one automatically.

    These attributes control how the model corresponds to the XML:

        * tagname:  the name of the tag representing this model
        * namespace:  the XML namespace in which this model lives

    These attributes control parsing/rendering behaviour:

        * namespace_prefix:  the prefix to use for rendering namespaced tags
        * ignore_unknown_elements:  ignore unknown elements when parsing
        * case_sensitive:    match tag/attr names case-sensitively
        * order_sensitive:   match child tags in order of field definition

    """

    _defaults = {
        "tagname": None,
        "namespace": None,
        "namespace_prefix": None,
        "ignore_unknown_elements": True,
        "case_sensitive": True,
        "order_sensitive": True,
    }

    def __init__(self, name: str, meta_attrs: Dict[str, str]) -> None:
        self.tagname = None
        for (attr, default) in self._defaults.items():
            setattr(self, attr, meta_attrs.get(attr, default))
        if self.tagname is None:
            self.tagname = name


def _meta_attributes(meta: Optional[type]) -> Dict[str, str]:
    """Extract attributes from a "meta" object."""
    meta_attrs = {}
    if meta:
        for attr in dir(meta):
            if not attr.startswith("_"):
                meta_attrs[attr] = getattr(meta, attr)
    return meta_attrs


class ModelMetaclass(type):
    """Metaclass for dexml.Model and subclasses.

    This metaclass is responsible for introspecting Model class definitions
    and setting up appropriate default behaviours.  For example, this metaclass
    sets a Model's default tagname to be equal to the declared class name.
    """

    instances_by_tagname: Dict[str, Any] = {}
    instances_by_classname: Dict[str, Any] = {}

    def __new__(mcls, name, bases, attrs):
        cls = super().__new__(mcls, name, bases, attrs)
        #  Don't do anything if it's not a subclass of Model
        parents = [b for b in bases if isinstance(b, ModelMetaclass)]
        if not parents:
            return cls
        #  Set up the cls.meta object, inheriting from base classes
        meta_attrs = {}
        for base in reversed(bases):
            if isinstance(base, ModelMetaclass) and hasattr(base, "meta"):
                meta_attrs.update(_meta_attributes(base.meta))
        meta_attrs.pop("tagname", None)
        meta_attrs.update(_meta_attributes(attrs.get("meta", None)))
        cls.meta = Meta(name, meta_attrs)
        #  Create ordered list of field objects, telling each about their
        #  name and containing class.  Inherit fields from base classes
        #  only if not overridden on the class itself.
        base_fields = {}
        for base in bases:
            if not isinstance(base, ModelMetaclass):
                continue
            for field in base._fields:
                if field.field_name not in base_fields:
                    field = copy.copy(field)
                    field.model_class = cls
                    base_fields[field.field_name] = field
        cls_fields = []
        for (name, value) in iter(attrs.items()):
            if isinstance(value, Field):
                base_fields.pop(name, None)
                value.field_name = name
                value.model_class = cls
                cls_fields.append(value)
        cls._fields = list(base_fields.values()) + cls_fields
        cls._fields.sort(key=lambda f: f._order_counter)
        #  Register the new class so we can find it by name later on
        tagname = (cls.meta.namespace, cls.meta.tagname)
        mcls.instances_by_tagname[tagname] = cls
        mcls.instances_by_classname[cls.__name__] = cls
        return cls

    @classmethod
    def find_class(mcls, tagname, namespace=None):
        """Find dexml.Model subclass for the given tagname and namespace."""
        try:
            return mcls.instances_by_tagname[(namespace, tagname)]
        except KeyError:
            if namespace is None:
                try:
                    return mcls.instances_by_classname[tagname]
                except KeyError:
                    pass
        return None



from xml.dom import minidom

from dexml import constants
from dexml.fields import field as dexml_field


class XmlNode(dexml_field.Field):
    tagname = None
    encoding = None

    class Arguments(dexml_field.Field.Arguments):
        tagname = None
        encoding = None

    def __set__(self, instance, value):
        if isinstance(value, str):
            if isinstance(value, str) and self.encoding:
                value = value.encode(self.encoding)
            doc = minidom.parseString(value)
            value = doc.documentElement
        if value is not None and value.namespaceURI is not None:
            nsattr = "xmlns"
            if value.prefix:
                nsattr = ":".join(
                    (
                        nsattr,
                        value.prefix,
                    )
                )
            value.attributes[nsattr] = value.namespaceURI
        return super().__set__(instance, value)

    def parse_child_node(self, obj, node):
        if self.tagname is None or self._check_tagname(node, self.tagname):
            self.__set__(obj, node)
            return constants.Status.PARSE_DONE
        return constants.Status.PARSE_SKIP

    @classmethod
    def render_children(cls, obj, val, nsmap):
        if val is not None:
            yield val.toxml()



from dexml import constants, exceptions
from dexml.fields import field as dexml_field
from dexml.fields import model as model_field


class Dict(dexml_field.Field):
    """Field subclass representing a dict of fields keyed by unique attribute value.

    This field corresponds to an indexed dict of other fields.  You would
    declare it like so:

      class MyObject(Model):
          name = fields.String(tagname = 'name')
          attr = fields.String(tagname = 'attr')

      class MyModel(Model):
          items = fields.Dict(fields.Model(MyObject), key = 'name')

    Corresponding to XML such as:

      <MyModel><MyObject><name>obj1</name><attr>val1</attr></MyObject></MyModel>


    The properties 'minlength' and 'maxlength' control the allowable size
    of the dict as in the List class.

    If 'unique' property is set to True, parsing will raise exception on
    non-unique key values.

    The 'dictclass' property controls the internal dict-like class used by
    the fielt.  By default it is the standard dict class.

    The 'tagname' property sets the 'wrapper' tag which acts as container
    for dict items, for example:

      from collections import defaultdict
      class MyObject(Model):
          name = fields.String()
          attr = fields.String()

      class MyDict(defaultdict):
          def __init__(self):
              super().__init__(MyObject)

      class MyModel(Model):
          objects = fields.Dict('MyObject', key = 'name',
                                tagname = 'dict', dictclass = MyDict)

      xml = '<MyModel><dict><MyObject name="obj1">'\
            <attr>val1</attr></MyObject></dict></MyModel>'
      mymodel = MyModel.parse(xml)
      obj2 = mymodel['obj2']
      print(obj2.name)
      print(mymodel.render(fragment = True))

    This wrapper tag is always rendered, even if the dict is empty.  It is
    transparently removed when parsing.
    """

    minlength: int
    maxlength: int
    unique: bool
    tagname: str
    dictclass: dict

    class Arguments(dexml_field.Field.Arguments):
        field = None
        minlength = None
        maxlength = None
        unique = False
        tagname = None
        dictclass = dict

    def __init__(self, field, key, **kwds):
        if isinstance(field, dexml_field.Field):
            kwds["field"] = field
        else:
            model_kwds = kwds.copy()
            if "tagname" in model_kwds:
                model_kwds.pop("tagname")
            kwds["field"] = model_field.Model(field, **model_kwds)
        super().__init__(**kwds)
        if not self.minlength and not self.tagname:
            self.required = False
        if self.minlength and not self.required:
            raise ValueError("Dict must be required if it has minlength")
        self.key = key

    def _get_field(self) -> dexml_field.Field:
        field = self.__dict__["field"]
        if not hasattr(field, "field_name"):
            field.field_name = self.field_name
        if not hasattr(field, "model_class"):
            field.model_class = self.model_class
        return field

    def _set_field(self, field) -> None:
        self.__dict__["field"] = field

    field = property(_get_field, _set_field)

    def __get__(self, instance, owner=None):
        val = super().__get__(instance, owner)
        if val is not None:
            return val

        class dictclass(self.dictclass):
            key = self.key

            def __setitem__(self, key, value):
                keyval = getattr(value, self.key)
                if keyval and keyval != key:
                    raise ValueError("Key field value does not match dict key")
                setattr(value, self.key, key)
                super().__setitem__(key, value)

        self.__set__(instance, dictclass())
        return self.__get__(instance, owner)

    def parse_child_node(self, obj, node):
        #  If our children are inside a grouping tag, parse
        #  that first.  The presence of this is indicated by
        #  setting an empty dict on the target object.
        if self.tagname:
            val = super().__get__(obj)
            if val is None:
                if node.nodeType != node.ELEMENT_NODE:
                    return constants.Status.PARSE_SKIP
                if node.tagName == self.tagname:
                    self.__get__(obj)
                    return constants.Status.PARSE_CHILDREN
                return constants.Status.PARSE_SKIP
        #  Now we just parse each child node.
        tmpobj = dexml_field.AttrBucket()
        res = self.field.parse_child_node(tmpobj, node)
        if res is constants.Status.PARSE_MORE:
            raise ValueError("items in a dict cannot return PARSE_MORE")
        if res is constants.Status.PARSE_DONE:
            items = self.__get__(obj)
            val = getattr(tmpobj, self.field_name)
            try:
                key = getattr(val, self.key)
            except AttributeError:
                raise exceptions.ParseError(
                    "Key field '%s' required but not found in dict value" % (self.key,)
                )
            if self.unique and key in items:
                raise exceptions.ParseError("Key '%s' already exists in dict" % (key,))
            items[key] = val
            return constants.Status.PARSE_MORE
        else:
            return constants.Status.PARSE_SKIP

    def parse_done(self, obj):
        items = self.__get__(obj)
        if self.minlength is not None and len(items) < self.minlength:
            raise exceptions.ParseError(f"Field '{self.field_name}': not enough items")
        if self.maxlength is not None and len(items) > self.maxlength:
            raise exceptions.ParseError("Field '{self.field_name}': too many items")

    def render_children(self, obj, items, nsmap):
        if self.minlength is not None and len(items) < self.minlength:
            raise exceptions.RenderError(f"Field '{self.field_name}': not enough items")
        if self.maxlength is not None and len(items) > self.maxlength:
            raise exceptions.RenderError("too many items")
        if self.tagname:
            children = "".join(
                data
                for item in items.values()
                for data in self.field.render_children(obj, item, nsmap)
            )
            if not children:
                if self.required:
                    yield f"<{self.tagname} />"
            else:
                yield children.join((f"<{self.tagname}>", f"</{self.tagname}>"))
        else:
            for item in items.values():
                for data in self.field.render_children(obj, item, nsmap):
                    yield data



from dexml import constants, exceptions
from dexml.fields import field as dexml_field
from dexml.fields import model as model_field


class List(dexml_field.Field):
    """Field subclass representing a list of fields.

    This field corresponds to a homogeneous list of other fields.  You would
    declare it like so:

      class MyModel(Model):
          items = fields.List(fields.String(tagname="item"))

    Corresponding to XML such as:

      <MyModel><item>one</item><item>two</item></MyModel>


    The properties 'minlength' and 'maxlength' control the allowable length
    of the list.

    The 'tagname' property sets an optional wrapper tag which acts as container
    for list items, for example:

      class MyModel(Model):
          items = fields.List(fields.String(tagname="item"),
                              tagname='list')

    Corresponding to XML such as:

      <MyModel><list><item>one</item><item>two</item></list></MyModel>

    This wrapper tag is always rendered, even if the list is empty.  It is
    transparently removed when parsing.
    """

    minlength: int
    maxlength: int
    tagname: str

    class Arguments(dexml_field.Field.Arguments):
        field = None
        minlength = None
        maxlength = None
        tagname = None

    def __init__(self, field, **kwds):
        if isinstance(field, dexml_field.Field):
            kwds["field"] = field
        else:
            model_kwds = kwds.copy()
            if "tagname" in model_kwds:
                model_kwds.pop("tagname")
            kwds["field"] = model_field.Model(field, **model_kwds)
        super().__init__(**kwds)
        if not self.minlength and not self.tagname:
            self.required = False
        if self.minlength and not self.required:
            raise ValueError("List must be required if it has minlength")

    def _get_field(self) -> dexml_field.Field:
        field = self.__dict__["field"]
        if not hasattr(field, "field_name"):
            field.field_name = self.field_name
        if not hasattr(field, "model_class"):
            field.model_class = self.model_class
        return field

    def _set_field(self, field) -> None:
        self.__dict__["field"] = field

    field = property(_get_field, _set_field)

    def __get__(self, instance, owner=None):
        val = super().__get__(instance, owner)
        if val is not None:
            return val
        self.__set__(instance, [])
        return self.__get__(instance, owner)

    def parse_child_node(self, obj, node):
        #  If our children are inside a grouping tag, parse
        #  that first.  The presence of this is indicated by
        #  setting the empty list on the target object.
        if self.tagname:
            val = super().__get__(obj)
            if val is None:
                if node.nodeType != node.ELEMENT_NODE:
                    return constants.Status.PARSE_SKIP
                elif node.tagName == self.tagname:
                    self.__set__(obj, [])
                    return constants.Status.PARSE_CHILDREN
                else:
                    return constants.Status.PARSE_SKIP
        #  Now we just parse each child node.
        tmpobj = dexml_field.AttrBucket()
        res = self.field.parse_child_node(tmpobj, node)
        if res is constants.Status.PARSE_MORE:
            raise ValueError("items in a list cannot return PARSE_MORE")
        if res is constants.Status.PARSE_DONE:
            items = self.__get__(obj)
            val = getattr(tmpobj, self.field_name)
            items.append(val)
            return constants.Status.PARSE_MORE
        else:
            return constants.Status.PARSE_SKIP

    def parse_done(self, obj):
        items = self.__get__(obj)
        if self.minlength is not None and len(items) < self.minlength:
            raise exceptions.ParseError(
                "Field '%s': not enough items" % (self.field_name,)
            )
        if self.maxlength is not None and len(items) > self.maxlength:
            raise exceptions.ParseError(
                "Field '%s': too many items" % (self.field_name,)
            )

    def render_children(self, obj, items, nsmap):
        #  Create a generator that yields child data chunks, and validates
        #  the number of items in the list as it goes.  It allows any
        #  iterable to be passed in, not just a list.
        def child_chunks():
            num_items = 0
            for item in items:
                num_items += 1
                if self.maxlength is not None and num_items > self.maxlength:
                    msg = f"Field '{self.field_name}': too many items"
                    raise exceptions.RenderError(msg)
                for data in self.field.render_children(obj, item, nsmap):
                    yield data
            if self.minlength is not None and num_items < self.minlength:
                msg = f"Field '{self.field_name}': not enough items"
                raise exceptions.RenderError(msg)

        chunks = child_chunks()
        #  Render each chunk, but suppress the wrapper tag if there's no data.
        try:
            data = next(chunks)
        except StopIteration:
            if self.tagname and self.required:
                yield f"<{self.tagname} />"
        else:
            if self.tagname:
                yield f"<{self.tagname}>"
            yield data
            for data in chunks:
                yield data
            if self.tagname:
                yield f"</{self.tagname}>"



"""

dexml.fields:  basic field type definitions for dexml
=====================================================

"""
import xml.dom.minidom
from typing import Any, Dict, Generator, List

from dexml import constants

#  Global counter tracking the order in which fields are declared.
_order_counter = 0


class AttrBucket:
    """A simple class used only to hold attributes."""


class Field(object):
    """Base class for all dexml Field classes.

    Field classes are responsible for parsing and rendering individual
    components to the XML.  They also act as descriptors on dexml Model
    instances, to get/set the corresponding properties.

    Each field instance will magically be given the following properties:

      * model_class:  the Model subclass to which it is attached
      * field_name:   the name under which is appears on that class

    The following methods are required for interaction with the parsing
    and rendering machinery:

      * parse_attributes:    parse info out of XML node attributes
      * parse_child_node:    parse into out of an XML child node
      * render_attributes:   render XML for node attributes
      * render_children:     render XML for child nodes

    """

    required: bool
    field_name: str
    model_class: Any

    class Arguments:
        required = True

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        """Default Field constructor.

        This constructor keeps track of the order in which Field instances
        are created, since this information can have semantic meaning in
        XML.  It also merges any keyword arguments with the defaults
        defined on the 'arguments' inner class, and assigned these attributes
        to the Field instance.
        """
        global _order_counter
        self._order_counter = _order_counter = _order_counter + 1
        args = self.__class__.Arguments
        for argnm in dir(args):
            if not argnm.startswith("__"):
                setattr(self, argnm, kwds.get(argnm, getattr(args, argnm)))

    def parse_attributes(
        self, obj: object, attrs: List[xml.dom.minidom.Attr]
    ) -> List[xml.dom.minidom.Attr]:
        """Parse any attributes for this field from the given list.

        This method will be called with the Model instance being parsed and
        a list of attribute nodes from its XML tag.  Any attributes of
        interest to this field should be processed, and a list of the unused
        attribute nodes returned.
        """
        return attrs

    def parse_child_node(
        self, obj: object, node: xml.dom.minidom.Element
    ) -> constants.Status:
        """Parse a child node for this field.

        This method will be called with the Model instance being parsed and
        the current child node of that model's XML tag.  There are three
        options for processing this node:

            * return PARSE_DONE, indicating that it was consumed and this
              field now has all the necessary data.
            * return PARSE_MORE, indicating that it was consumed but this
              field will accept more nodes.
            * return PARSE_SKIP, indicating that it was not consumed by
              this field.

        Any other return value will be taken as a parse error.
        """
        return constants.Status.PARSE_SKIP

    def parse_done(self, obj: object) -> None:
        """Finalize parsing for the given object.

        This method is called as a simple indicator that no more data will
        be forthcoming.  No return value is expected.
        """

    def render_attributes(
        self, obj: object, val: Any, nsmap: Dict[str, Any]
    ) -> Generator[str, None, None]:
        """Render any attributes that this field manages."""
        return
        yield

    def render_children(
        self, obj: object, val: Any, nsmap: Dict[str, Any]
    ) -> Generator[str, None, None]:
        """Render any child nodes that this field manages."""
        return
        yield

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return instance.__dict__.get(self.field_name)

    def __set__(self, instance, value):
        instance.__dict__[self.field_name] = value

    def _check_tagname(self, node, tagname):
        if node.nodeType != node.ELEMENT_NODE:
            return False
        if isinstance(tagname, str):
            if node.localName != tagname:
                return False
            if node.namespaceURI:
                if node.namespaceURI != self.model_class.meta.namespace:
                    return False
        else:
            (tagns, tagname) = tagname
            if node.localName != tagname:
                return False
            if node.namespaceURI != tagns:
                return False
        return True



from dexml.fields.choice import Choice
from dexml.fields.dict import Dict
from dexml.fields.field import Field
from dexml.fields.list import List
from dexml.fields.model import Model
from dexml.fields.values import CDATA, Boolean, Float, Integer, String
from dexml.fields.xmlnode import XmlNode

__all__ = [
    "Field",
    "String",
    "Boolean",
    "Integer",
    "CDATA",
    "Float",
    "Choice",
    "Model",
    "List",
    "Dict",
    "XmlNode",
]



class Error(Exception):
    """Base exception class for the dexml module."""


class ParseError(Error):
    """Exception raised when XML could not be parsed into objects."""


class RenderError(Error):
    """Exception raised when object could not be rendered into XML."""


class XmlError(Error):
    """Exception raised to encapsulate errors from underlying XML parser."""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dexml', 'dexml.cli', 'dexml.fields']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'dexml-python3',
    'version': '0.1.1',
    'description': 'A dead-simple xml object mapper for Python',
    'long_description': '# dexml: An Object-XML Mapper for Python\n\nThis is a maintained version of the original `dexml` Python package, intended for Python 3 use only.\n\nOther than providing a maintained version that supports modern Python 3, other future goals include:\n - support with Python dataclasses and `pydantic` models\n - support for auto-generating models based on example XML payloads\n - keep backwards compatibility with original `dexml` package where possible\n\nBefore refactoring, the original code was sourced from both:\n - https://github.com/realrunner/dexml\n - https://github.com/rfk/dexml\n\n## Documentation\n\nLet\'s face it: xml is a fact of modern life. I\'d even go so far as to say\nthat it\'s *good* at what it does. But that doesn\'t mean it\'s easy to work\nwith and it doesn\'t mean that we have to like it. Most of the time, XML\njust needs to get out of the way and let you do some actual work instead\nof writing code to traverse and manipulate yet another DOM.\n\nThe dexml module takes the obvious mapping between XML tags and Python objects\nand lets you capture that as cleanly as possible. Loosely inspired by Django\'s\nORM, you write simple class definitions to define the expected structure of\nyour XML document. Like so::\n\n```\n  >>> import dexml\n  >>> from dexml import fields\n  >>> class Person(dexml.Model):\n  ...   name = fields.String()\n  ...   age = fields.Integer(tagname=\'age\')\n```\n\nThen you can parse an XML document into an object like this::\n\n```\n  >>> p = Person.parse("<Person name=\'Foo McBar\'><age>42</age></Person>")\n  >>> p.name\n  u\'Foo McBar\'\n  >>> p.age\n  42\n```\n\nAnd you can render an object into an XML document like this::\n\n```\n  >>> p = Person(name="Handsome B. Wonderful",age=36)\n  >>> p.render()\n  \'<?xml version="1.0" ?><Person name="Handsome B. Wonderful"><age>36</age></Person>\'\n```\n\nMalformed documents will raise a ParseError::\n\n```\n  >>> p = Person.parse("<Person><age>92</age></Person>")\n  Traceback (most recent call last):\n      ...\n  ParseError: required field not found: \'name\'\n```\n\nOf course, it gets more interesting when you nest Model definitions, like this::\n\n```\n  >>> class Group(dexml.Model):\n  ...   name = fields.String(attrname="name")\n  ...   members = fields.List(Person)\n  ...\n  >>> g = Group(name="Monty Python")\n  >>> g.members.append(Person(name="John Cleese",age=69))\n  >>> g.members.append(Person(name="Terry Jones",age=67))\n  >>> g.render(fragment=True)\n  \'<Group name="Monty Python"><Person name="John Cleese"><age>69</age></Person><Person name="Terry Jones"><age>67</age></Person></Group>\'\n```\n\nThere\'s support for XML namespaces, default field values, case-insensitive\nparsing, and more fun stuff. Check out the documentation on the following\nclasses for more details:\n\n```\n  :Model:  the base class for objects that map into XML\n  :Field:  the base class for individual model fields\n  :Meta:   meta-information about how to parse/render a model\n```\n\n## Auto-Generating Models\n\nYou can generate Python code with `dexml` models from XML using the following command:\n\n```\ncat file.xml | python -m dexml > model.py\n```\n\nOnly the `Model`, `String`, `Integer`, `Float`, and `Boolean` fields are currently supported.\n\nGeneration is not intended to be 100% perfect but as a starting point for manually creating large models.\n\n## Development\n\nThe project uses [poetry](https://python-poetry.org/) to manage dependencies, virtual environments, and publishing.\n\nIt also uses `pre-commit` to provide some standard git hook checks.\n\nThe `Makefile` has some shortcuts for common operations. After cloning, the following steps can be taken:\n1. Run `make restore`\n2. Optionally enter virtual environment with `poetry shell`\n3. Run `make test`\n4. Run `make build` to create a distribution\n\n## Contributing\n\nThis is an early stage project but pull requests are welcome.\n',
    'author': 'Patrick Withams',
    'author_email': 'pwithams@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/pwithams/dexml-python3',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from dexml import constants, exceptions
from dexml.fields import field as dexml_field
from dexml.meta import ModelMetaclass


class Model(dexml_field.Field):
    """Field subclass referencing another Model instance.

    This field sublcass allows Models to contain other Models recursively.
    The first argument to the field constructor must be either a Model
    class, or the name or tagname of a Model class.
    """

    class Arguments(dexml_field.Field.Arguments):
        type = None

    def __init__(self, type=None, **kwds):
        kwds["type"] = type
        self.tagname = kwds.get("tagname", None)
        super().__init__(**kwds)

    def _get_type(self):
        return self.__dict__.get("type")

    def _set_type(self, value):
        if value is not None:
            self.__dict__["type"] = value

    type = property(_get_type, _set_type)

    def __set__(self, instance, value):
        typeclass = self.typeclass
        if value and not isinstance(value, typeclass):
            raise ValueError(
                "Invalid value type %s. Model field requires %s instance"
                % (value.__class__.__name__, typeclass.__name__)
            )
        super().__set__(instance, value)

    @property
    def typeclass(self):
        try:
            return self.__dict__["typeclass"]
        except KeyError:
            self.__dict__["typeclass"] = self._load_typeclass()
            return self.__dict__["typeclass"]

    def _load_typeclass(self):
        typ = self.type
        if isinstance(typ, ModelMetaclass):
            return typ
        if typ is None:
            typ = self.field_name
        typeclass = None
        if isinstance(typ, str):
            if self.model_class.meta.namespace:
                ns = self.model_class.meta.namespace
                typeclass = ModelMetaclass.find_class(typ, ns)
            if typeclass is None:
                typeclass = ModelMetaclass.find_class(typ, None)
            if typeclass is None:
                raise ValueError("Unknown Model class: %s" % (typ,))
        else:
            (ns, typ) = typ
            if isinstance(typ, ModelMetaclass):
                return typ
            typeclass = ModelMetaclass.find_class(typ, ns)
            if typeclass is None:
                raise ValueError("Unknown Model class: (%s,%s)" % (ns, typ))
        return typeclass

    def parse_child_node(self, obj, node):
        typeclass = self.typeclass
        try:
            typeclass.validate_xml_node(node, self.tagname)
        except exceptions.ParseError:
            return constants.Status.PARSE_SKIP
        else:
            inst = typeclass.parse(node, self.tagname)
            self.__set__(obj, inst)
            return constants.Status.PARSE_DONE

    def render_attributes(self, obj, val, nsmap):
        return []

    def render_children(self, obj, val, nsmap):
        if val is not None:
            for data in val._render(nsmap, self.tagname):
                yield data



import io
import sys
import xml.etree.ElementTree as xml
from dataclasses import dataclass
from typing import List, Union

TABSIZE = 4
TAB = " " * TABSIZE


def get_type(value: str) -> str:
    if value.isdigit():
        return "Integer"
    if value.replace(".", "").isdigit():
        return "Float"
    if value in ["true", "false"]:
        return "Boolean"
    return "String"


def to_snake_case(name: str) -> str:
    new_name = []
    prev_char = None
    for char in name:
        if (
            char.isupper()
            and prev_char
            and prev_char != "_"
            and not prev_char.isupper()
        ):
            new_name.append("_")
        if (
            char.islower()
            and prev_char
            and prev_char.isupper()
            and len(new_name) > 1
            and new_name[-2] != "_"
        ):
            new_name[-1] = "_"
            new_name.append(prev_char.lower())
        if not char.isalpha():
            new_name.append("_")
        else:
            new_name.append(char.lower())
        prev_char = char
    result = "".join(new_name)
    return result.strip("_")


def to_pascal_case(name: str) -> str:
    new_name = []
    cap_next = True
    prev_char_lower = False
    for char in name:
        if char == "_":
            cap_next = True
            continue
        if cap_next or (prev_char_lower and char.isupper()):
            new_name.append(char.upper())
            cap_next = False
        else:
            new_name.append(char.lower())
        prev_char_lower = char.islower()
    return "".join(new_name)


@dataclass
class DexmlAttribute:
    tagname: str
    type: str = "String"

    def __str__(self) -> str:
        name = to_snake_case(self.tagname)
        if name == self.tagname:
            return f"{self.tagname} = fields.{self.type}()"
        return f'{name} = fields.{self.type}(tagname="{self.tagname}")'


@dataclass
class DexmlModel:
    tagname: str
    attributes: List[DexmlAttribute]
    models: List["DexmlModel"]
    level: int

    def get_as_attribute(self) -> str:
        attr_name = to_snake_case(self.tagname)
        cls_name = to_pascal_case(self.tagname)
        return f"{attr_name} = fields.Model({cls_name})"

    def __str__(self) -> str:
        name = to_pascal_case(self.tagname)
        data = [
            f"class {name}(dexml.Model):",
        ]
        if name != self.tagname:
            data.append(f"{TAB}class meta:")
            data.append(f'{TAB}{TAB}tagname = "{self.tagname}"')
        attribute_lines = [TAB + str(attr) for attr in self.attributes]
        model_lines = [TAB + str(model.get_as_attribute()) for model in self.models]
        # model_definitions = [str(model) for model in self.models]
        return "\n".join(data + attribute_lines + model_lines)


def extract_models(
    xml_element: xml.ElementTree, level: int, all_models: List[DexmlModel]
) -> Union[DexmlAttribute, DexmlModel]:
    children_count = 0
    for xml_element_child in xml_element:
        children_count += 1

    if children_count == 0:
        return DexmlAttribute(xml_element.tag, type=get_type(xml_element.text))

    dexml_model = DexmlModel(xml_element.tag, [], [], level)

    for xml_element_child in xml_element:
        result = extract_models(xml_element_child, level + 1, all_models)
        space = "   " * (level + 1)
        if isinstance(result, DexmlAttribute):
            dexml_model.attributes.append(result)
        else:
            dexml_model.models.append(result)
            all_models.append(result)

    return dexml_model


def parse(xml_string):
    data = xml.parse(io.StringIO(xml_string))
    all_models = []
    top_level_model = extract_models(data.getroot(), 0, all_models)
    models = [top_level_model, *all_models]
    models = sorted(models, key=lambda model: model.level, reverse=True)
    result = []
    result.append("import dexml")
    result.append("from dexml import fields")
    result.append("")
    generated = {}
    for model in models:
        model_str = str(model)
        if model_str in generated:
            continue
        result.append("")
        result.append(model_str)
        result.append("")
        generated[model_str] = True

    return "\n".join(result).rstrip("\n")


def main():
    xml_lines = []
    for line in sys.stdin:
        xml_lines.append(line)

    xml_string = "".join(xml_lines)
    result = parse(xml_string)
    print(result)
